import openai
from openai._types import FileTypes


def transcribe_with_api(audio_file: FileTypes, prompt: str | None = None):
    """
    Transcribe an audio file using the OpenAI Whisper API
    """
    transcript = openai.audio.transcriptions.create(
        model="whisper-1",
        file=open(audio_file, "rb"),
        response_format="verbose_json",
        timestamp_granularities=["segment", "word"],
        prompt=prompt,
    )

    # Add space to beginning of words
    # to match local Whisper format
    for word in transcript.words:
        word["word"] = " " + word["word"]

    # Return response in same format
    # as local Whisper format
    return [
        {
            "start": transcript.segments[0]["start"],
            "end": transcript.segments[-1]["end"],
            "words": transcript.words,
        }
    ]


def transcribe_locally(
    audio_file: str, model_name: str = "base", prompt: str | None = None
):
    """
    Transcribe an audio file using the local Whisper package
    (https://pypi.org/project/openai-whisper/)
    """
    import whisper

    if model_name not in whisper.available_models():
        raise ValueError(f"Model {model_name} not found")

    model = whisper.load_model(model_name)

    transcription = model.transcribe(
        audio=audio_file,
        word_timestamps=True,
        fp16=False,
        initial_prompt=prompt,
    )

    return transcription["segments"]



import tempfile

import numpy
from moviepy.editor import CompositeVideoClip, ImageClip, TextClip, VideoClip
from PIL import Image, ImageFilter, ImageFont

text_cache = {}
shadow_cache = {}


class Character:
    def __init__(self, text, color=None):
        self.text = text
        self.color = color

    def set_color(self, color):
        self.color = color


class Word:
    def __init__(self, word, color=None):
        self.word = word
        self.color = color
        self.characters = []

        for char in word:
            self.characters.append(Character(char, color))

    def set_color(self, color):
        self.color = color
        for char in self.characters:
            char.set_color(color)


class TextClipEx(TextClip):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.text = kwargs["txt"]


def moviepy_to_pillow(clip) -> Image:
    temp_file = tempfile.NamedTemporaryFile(suffix=".png").name
    clip.save_frame(temp_file)
    image = Image.open(temp_file)
    return image


def get_text_size(text, fontsize, font, stroke_width):
    text_clip = create_text(
        text, fontsize=fontsize, color="white", font=font, stroke_width=stroke_width
    )
    return text_clip.size


def get_text_size_ex(text, font: tuple[str, str], fontsize, stroke_width):
    text_clip = create_text_ex(
        text, fontsize=fontsize, color="white", font=font, stroke_width=stroke_width
    )
    return text_clip.size


def blur_text_clip(text_clip, blur_radius: int) -> VideoClip:
    # Convert TextClip to a PIL image
    pil_img = moviepy_to_pillow(text_clip)

    # Offset blur to make it centered
    offset = int(blur_radius * 0.6)

    # Add empty space around text for blur
    pil_img_padded = Image.new(
        "RGBA", (pil_img.width + blur_radius * 3, pil_img.height + blur_radius * 3)
    )
    pil_img_padded.paste(
        moviepy_to_pillow(text_clip), (blur_radius + offset, blur_radius + offset)
    )

    # Create a blurred version of the text
    pil_img_padded = pil_img_padded.filter(ImageFilter.GaussianBlur(radius=blur_radius))

    text_clip = ImageClip(numpy.array(pil_img_padded))
    text_clip = text_clip.set_duration(text_clip.duration)

    return text_clip


def create_text(
    text: str,
    fontsize: int,
    color: str,
    font: str,
    bg_color: str = "transparent",
    blur_radius: int = 0,
    opacity: float = 1.0,
    stroke_color: str | None = None,
    stroke_width: int = 1,
    kerning: float = 0.0,
) -> VideoClip:
    global text_cache

    arg_hash = hash(
        (
            text,
            fontsize,
            color,
            font,
            bg_color,
            blur_radius,
            opacity,
            stroke_color,
            stroke_width,
            kerning,
        )
    )

    if arg_hash in text_cache:
        return text_cache[arg_hash].copy()

    text_clip = TextClipEx(
        txt=text,
        fontsize=fontsize,
        color=color,
        bg_color=bg_color,
        font=font,
        stroke_color=stroke_color,
        stroke_width=stroke_width,
        kerning=kerning,
    )

    text_clip = text_clip.set_opacity(opacity)

    if blur_radius:
        text_clip = blur_text_clip(text_clip, blur_radius)

    text_cache[arg_hash] = text_clip.copy()

    return text_clip


def create_text_chars(
    text: list[Word] | list[Character],
    fontsize,
    color,
    font,
    bg_color="transparent",
    blur_radius: int = 0,
    opacity=1,
    stroke_color=None,
    stroke_width=1,
    add_space_between_words=True,
) -> list[VideoClip]:
    # Create a clip for each character
    clips = []
    for i, item in enumerate(text):
        if isinstance(item, Word):
            chars = item.characters
            if add_space_between_words and i < len(text) - 1:
                chars.append(Character(" ", item.color))
        else:
            chars = [item]

        for char in chars:
            clip = create_text(
                char.text,
                fontsize,
                char.color or color,
                font,
                bg_color,
                blur_radius,
                opacity,
                stroke_color,
                stroke_width,
            )
            clips.append(clip)

    return clips


def create_composite_text(
    text_clips: list[VideoClip], font, font_size
) -> CompositeVideoClip:
    clips = []

    font = ImageFont.truetype(font, font_size)
    scale_factor = 3.012  # factor to convert Pillow to MoviePy width

    full_width = 0
    for clip in text_clips[:-1]:
        width = font.getlength(clip.text) * scale_factor
        full_width += width

    full_width += text_clips[-1].size[0]
    offset_x = 0

    for clip in text_clips:
        clip.size = (int(full_width), clip.size[1])
        clip = clip.set_position((int(offset_x), 0))
        width = font.getlength(clip.text)
        offset_x += width * scale_factor
        clips.append(clip)

    return CompositeVideoClip(clips)


def str_to_charlist(text: str) -> list[Character]:
    return [Character(char) for char in text]


def create_text_ex(
    text: list[Word] | list[Character] | str,
    fontsize,
    color,
    font: tuple[str, str],
    bg_color="transparent",
    blur_radius: int = 0,
    opacity=1,
    stroke_color=None,
    stroke_width=1,
    kerning=0,
) -> CompositeVideoClip:
    if isinstance(text, str):
        text = str_to_charlist(text)
    text_clips = create_text_chars(
        text,
        fontsize,
        color,
        font[0],
        bg_color,
        blur_radius,
        opacity,
        stroke_color,
        stroke_width,
    )
    return create_composite_text(text_clips, font[1], fontsize // 3)


def create_shadow(
    text: str,
    font_size: int,
    font: tuple[str, str],
    blur_radius: float,
    opacity: float = 1.0,
):
    global shadow_cache

    arg_hash = hash((text, font_size, font, blur_radius, opacity))

    if arg_hash in shadow_cache:
        return shadow_cache[arg_hash].copy()

    shadow = create_text_ex(text, font_size, "black", font, opacity=opacity)
    shadow = blur_text_clip(shadow, int(font_size * blur_radius))

    shadow_cache[arg_hash] = shadow.copy()

    return shadow



import os
import tempfile
import time

import ffmpeg
from moviepy.config import change_settings
from moviepy.editor import CompositeVideoClip, VideoFileClip

from . import segment_parser, transcriber
from .__version__ import __version__
from .errors import UtilityNotFoundError
from .text_drawer import Word, create_shadow, create_text_ex, get_text_size_ex
from .utils import (
    _get_font_path,
    ffmpeg_binary,
    ffmpeg_installed,
    imagemagick_binary,
    is_local_transcription_available,
)

__all__ = [
    "__version__",
    "add_captions",
    "ffmpeg_installed",
    "imagemagick_binary",
    "is_local_transcription_available",
]

lines_cache = {}

if not ffmpeg_installed():
    raise UtilityNotFoundError("FFmpeg")
else:
    change_settings({"FFMPEG_BINARY": ffmpeg_binary()})
    os.environ["IMAGEIO_FFMPEG_EXE"] = ffmpeg_binary()

if imagemagick_binary() == "unset":
    raise UtilityNotFoundError("ImageMagick")
elif not os.path.exists(imagemagick_binary()):
    raise FileNotFoundError(f"ImageMagick binary '{imagemagick_binary()}' not found")
else:
    change_settings({"IMAGEMAGICK_BINARY": imagemagick_binary()})


def fits_frame(line_count, font: tuple[str, str], font_size, stroke_width, frame_width):
    def fit_function(text):
        lines = calculate_lines(text, font, font_size, stroke_width, frame_width)
        return len(lines["lines"]) <= line_count

    return fit_function


def calculate_lines(
    text,
    font: tuple[str, str],
    font_size,
    stroke_width,
    frame_width,
    verbose: bool = False,
):
    global lines_cache

    arg_hash = hash((text, font, font_size, stroke_width, frame_width))

    if arg_hash in lines_cache:
        return lines_cache[arg_hash]

    lines = []

    line_to_draw = None
    line = ""
    words = text.split()
    word_index = 0
    total_height = 0
    while word_index < len(words):
        word = words[word_index]
        line += word + " "
        text_size = get_text_size_ex(line.strip(), font, font_size, stroke_width)
        text_width = text_size[0]
        line_height = text_size[1]

        if text_width < frame_width:
            line_to_draw = {
                "text": line.strip(),
                "height": line_height,
            }
            word_index += 1
        else:
            if not line_to_draw:
                if verbose:
                    print(f"NOTICE: Word '{line.strip()}' is too long for the frame!")
                line_to_draw = {
                    "text": line.strip(),
                    "height": line_height,
                }
                word_index += 1

            lines.append(line_to_draw)
            total_height += line_height
            line_to_draw = None
            line = ""

    if line_to_draw:
        lines.append(line_to_draw)
        total_height += line_height

    data = {
        "lines": lines,
        "height": total_height,
    }

    lines_cache[arg_hash] = data

    return data


def add_captions(
    video_file: str,
    output_file: str = "with_transcript.mp4",
    font_path: str = "Bangers-Regular.ttf",
    font_size: int = 100,
    font_color: str = "white",
    stroke_width: int = 3,
    stroke_color: str = "black",
    highlight_current_word: bool = True,
    highlight_color: str = "yellow",
    line_count: int = 2,
    fit_function=None,
    rel_width: float = 0.6,
    rel_height_pos: float = 0.5,  # TODO: Implement this
    shadow_strength: float = 1.0,
    shadow_blur: float = 0.1,
    verbose: bool = False,
    initial_prompt: str | None = None,
    segments=None,
    model_name: str = "base",
    use_local_whisper: str | bool = "auto",
    temp_audiofile: str | None = None,
):
    _start_time = time.time()

    font_path, injected_font_name = _get_font_path(font_path)

    if verbose:
        print("Extracting audio...")

    temp_audiofile = temp_audiofile or tempfile.NamedTemporaryFile(suffix=".wav").name
    ffmpeg.input(video_file).output(
        temp_audiofile, loglevel="info" if verbose else "quiet"
    ).run(overwrite_output=True)

    if segments is None:
        if verbose:
            print("Transcribing audio...")

        if use_local_whisper == "auto":
            use_local_whisper = is_local_transcription_available(verbose)

        if use_local_whisper:
            segments = transcriber.transcribe_locally(
                audio_file=temp_audiofile, prompt=initial_prompt, model_name=model_name
            )
        else:
            segments = transcriber.transcribe_with_api(
                audio_file=temp_audiofile, prompt=initial_prompt
            )

    if verbose:
        print("Generating video elements...")

    # Open the video file
    video = VideoFileClip(video_file)
    text_bbox_width = video.w * rel_width
    clips = [video]

    captions = segment_parser.parse(
        segments=segments,
        fit_function=(
            fit_function
            if fit_function
            else fits_frame(
                line_count,
                (injected_font_name, font_path),
                font_size,
                stroke_width,
                text_bbox_width,
            )
        ),
    )

    for caption in captions:
        captions_to_draw = []
        if highlight_current_word:
            for i, word in enumerate(caption["words"]):
                if i + 1 < len(caption["words"]):
                    end = caption["words"][i + 1]["start"]
                else:
                    end = word["end"]

                captions_to_draw.append(
                    {
                        "text": caption["text"],
                        "start": word["start"],
                        "end": end,
                    }
                )
        else:
            captions_to_draw.append(caption)

        for current_index, caption in enumerate(captions_to_draw):
            line_data = calculate_lines(
                caption["text"],
                (injected_font_name, font_path),
                font_size,
                stroke_width,
                text_bbox_width,
                verbose=verbose,
            )

            text_y_offset = video.h * (1 - rel_height_pos) - line_data["height"] // 2
            index = 0
            for line in line_data["lines"]:
                pos = ("center", text_y_offset)

                words = line["text"].split()
                word_list = []
                for w in words:
                    word_obj = Word(w)
                    if highlight_current_word and index == current_index:
                        word_obj.set_color(highlight_color)
                    index += 1
                    word_list.append(word_obj)

                # Create shadow
                shadow_left = shadow_strength
                while shadow_left >= 1:
                    shadow_left -= 1
                    shadow = create_shadow(
                        line["text"],
                        font_size,
                        (injected_font_name, font_path),
                        shadow_blur,
                        opacity=1,
                    )
                    shadow = shadow.set_start(caption["start"])
                    shadow = shadow.set_duration(caption["end"] - caption["start"])
                    shadow = shadow.set_position(pos)
                    clips.append(shadow)

                if shadow_left > 0:
                    shadow = create_shadow(
                        line["text"],
                        font_size,
                        (injected_font_name, font_path),
                        shadow_blur,
                        opacity=shadow_left,
                    )
                    shadow = shadow.set_start(caption["start"])
                    shadow = shadow.set_duration(caption["end"] - caption["start"])
                    shadow = shadow.set_position(pos)
                    clips.append(shadow)

                # Create text
                text = create_text_ex(
                    word_list,
                    font_size,
                    font_color,
                    (injected_font_name, font_path),
                    stroke_color=stroke_color,
                    stroke_width=stroke_width,
                )
                text = text.set_start(caption["start"])
                text = text.set_duration(caption["end"] - caption["start"])
                text = text.set_position(pos)
                clips.append(text)

                text_y_offset += line["height"]

    end_time = time.time()
    generation_time = end_time - _start_time

    if verbose:
        print(
            f"Generated in {generation_time//60:02.0f}:{generation_time%60:02.0f} ({len(clips)} clips)"
        )

    if verbose:
        print("Rendering video...")

    video_with_text = CompositeVideoClip(clips)

    video_with_text.write_videofile(
        filename=output_file,
        codec="libx264",
        fps=video.fps,
        logger="bar" if verbose else None,
        temp_audiofile=temp_audiofile,
    )

    end_time = time.time()
    total_time = end_time - _start_time
    render_time = total_time - generation_time

    if verbose:
        print(f"Generated in {generation_time//60:02.0f}:{generation_time%60:02.0f}")
        print(f"Rendered in {render_time//60:02.0f}:{render_time%60:02.0f}")
        print(f"Done in {total_time//60:02.0f}:{total_time%60:02.0f}")



#!/usr/bin/env python3

import re
from pathlib import Path
from typing import Annotated, Optional

import typer
import typer.core

from captametropolis import add_captions
from captametropolis.__version__ import __version__
from captametropolis.utils import is_font_registered, register_font, unregister_font


class AliasGroup(typer.core.TyperGroup):
    _CMD_SPLIT_P = r"[,| ?\/]"

    def get_command(self, ctx, cmd_name):
        cmd_name = self._group_cmd_name(cmd_name)
        return super().get_command(ctx, cmd_name)

    def _group_cmd_name(self, default_name):
        for cmd in self.commands.values():
            if cmd.name and default_name in re.split(self._CMD_SPLIT_P, cmd.name):
                return cmd.name
        return default_name


app: typer.Typer = typer.Typer(
    name="Captametropolis",
    help=":sparkles: An [italic]awesome[/italic] [orange1]CLI tool[/orange1] to add captions to your videos. :movie_camera: :rocket:",
    rich_markup_mode="rich",
    epilog=f"Made with [red]:red_heart:[/red]  and :muscle: by [cyan link=https://github.com/AppSolves]AppSolves[/cyan link] | [blue link=https://github.com/AppSolves/captametropolis]GitHub[/blue link] | [green]v{__version__}[/green] from [yellow link=https://pypi.org/project/captametropolis/]PyPI[/yellow link]",
    cls=AliasGroup,
    context_settings={
        "help_option_names": ["-h", "--help", "-?"],
    },
)


@app.command(
    name="add_caption",
    help="Add captions to a video file.",
)
def create(
    video_file: Annotated[
        Path,
        typer.Argument(
            ...,
            help="The path to the video file to which captions will be added.",
            show_default=False,
        ),
    ],
    output_file: Annotated[
        Path,
        typer.Argument(
            ...,
            help="The path to the output video file with captions added.",
            show_default=False,
        ),
    ],
    font_path: Annotated[
        Optional[Path],
        typer.Option(
            ...,
            "--font-path",
            "-fp",
            help="The path to the font file to be used for the captions.",
        ),
    ] = "Bangers-Regular.ttf",
    font_size: Annotated[
        Optional[int],
        typer.Option(
            ...,
            "--font-size",
            "-fs",
            help="The size of the font to be used for the captions.",
        ),
    ] = 100,
    font_color: Annotated[
        Optional[str],
        typer.Option(
            ...,
            "--font-color",
            "-fc",
            help="The color of the font to be used for the captions.",
        ),
    ] = "white",
    stroke_width: Annotated[
        Optional[int],
        typer.Option(
            ...,
            "--stroke-width",
            "-sw",
            help="The width of the stroke to be used for the captions.",
        ),
    ] = 3,
    stroke_color: Annotated[
        Optional[str],
        typer.Option(
            ...,
            "--stroke-color",
            "-sc",
            help="The color of the stroke to be used for the captions.",
        ),
    ] = "black",
    highlight_current_word: Annotated[
        Optional[bool],
        typer.Option(
            ...,
            "--highlight-current-word",
            "-hcw",
            help="Highlight the current word being spoken in the captions.",
        ),
    ] = True,
    highlight_color: Annotated[
        Optional[str],
        typer.Option(
            ...,
            "--highlight-color",
            "-hc",
            help="The color of the highlight to be used for the current word.",
        ),
    ] = "yellow",
    line_count: Annotated[
        Optional[int],
        typer.Option(
            ...,
            "--line-count",
            "-lc",
            help="The number of lines to be displayed in the captions.",
        ),
    ] = 2,
    rel_width: Annotated[
        Optional[float],
        typer.Option(
            ...,
            "--rel-width",
            "-rw",
            help="The relative width of the captions to the video frame.",
        ),
    ] = 0.6,
    rel_height_pos: Annotated[
        Optional[float],
        typer.Option(
            ...,
            "--rel-height-pos",
            "-rhp",
            help="The relative height position of the captions in the video frame.",
        ),
    ] = 0.5,
    shadow_strength: Annotated[
        Optional[float],
        typer.Option(
            ...,
            "--shadow-strength",
            "-ss",
            help="The strength of the shadow to be used for the captions.",
        ),
    ] = 1.0,
    shadow_blur: Annotated[
        Optional[float],
        typer.Option(
            ...,
            "--shadow-blur",
            "-sb",
            help="The blur of the shadow to be used for the captions.",
        ),
    ] = 0.1,
    verbose: Annotated[
        Optional[bool],
        typer.Option(
            ...,
            "--verbose",
            "-v",
            help="Show verbose output.",
        ),
    ] = False,
    initial_prompt: Annotated[
        Optional[str],
        typer.Option(
            ...,
            "--initial-prompt",
            "-ip",
            help="The initial prompt to be passed to Whisper.",
            show_default=False,
        ),
    ] = None,
    model_name: Annotated[
        Optional[str],
        typer.Option(
            ...,
            "--model-name",
            "-mn",
            help="The name of the model to be used for Whisper.",
        ),
    ] = "base",
    use_local_whisper: Annotated[
        Optional[bool],
        typer.Option(
            ...,
            "--use-local-whisper",
            "-ulw",
            help="Use the local Whisper model if available.",
        ),
    ] = True,
    temp_audiofile: Annotated[
        Optional[Path],
        typer.Option(
            ...,
            "--temp-audiofile",
            "-ta",
            help="The path to the temporary audio file to be used for Whisper.",
            show_default=False,
        ),
    ] = None,
):
    add_captions(
        video_file=str(video_file.resolve()),
        output_file=str(output_file.resolve()),
        font_path=str(font_path.resolve()),
        font_size=font_size,
        font_color=font_color,
        stroke_width=stroke_width,
        stroke_color=stroke_color,
        highlight_current_word=highlight_current_word,
        highlight_color=highlight_color,
        line_count=line_count,
        rel_width=rel_width,
        rel_height_pos=rel_height_pos,
        shadow_strength=shadow_strength,
        shadow_blur=shadow_blur,
        verbose=verbose,
        initial_prompt=initial_prompt,
        model_name=model_name,
        use_local_whisper=use_local_whisper,
        temp_audiofile=str(temp_audiofile.resolve()) if temp_audiofile else None,
    )


@app.command(
    name="register_font",
    help="Register a font file to be used for captions.",
)
def register_font_cmd(
    font_path: Annotated[
        Path,
        typer.Argument(
            ...,
            help="The path to the font file to be registered.",
            show_default=False,
        ),
    ],
    quiet_run: Annotated[
        Optional[bool],
        typer.Option(
            ...,
            "--quiet-run",
            "-qr",
            help="Run the command quietly without any output or console.",
        ),
    ] = False,
    verbose: Annotated[
        Optional[bool],
        typer.Option(
            ...,
            "--verbose",
            "-v",
            help="Show verbose output.",
        ),
    ] = False,
):
    registered_font_name = register_font(
        str(font_path.resolve()),
        quiet_run=quiet_run,
        verbose=verbose,
    )
    typer.echo(f"Registered font: {registered_font_name}")


@app.command(
    name="unregister_font",
    help="Unregister a font file from being used for captions.",
)
def unregister_font_cmd(
    font_path_or_name: Annotated[
        str,
        typer.Argument(
            ...,
            help="The path of the font file or the font's name to be unregistered.",
            show_default=False,
        ),
    ],
    quiet_run: Annotated[
        Optional[bool],
        typer.Option(
            ...,
            "--quiet-run",
            "-qr",
            help="Run the command quietly without any output or console.",
        ),
    ] = False,
    verbose: Annotated[
        Optional[bool],
        typer.Option(
            ...,
            "--verbose",
            "-v",
            help="Show verbose output.",
        ),
    ] = False,
):
    result = unregister_font(
        font_path_or_name,
        quiet_run=quiet_run,
        verbose=verbose,
    )
    if result:
        typer.echo(f"Unregistered font: {font_path_or_name}")
    else:
        typer.echo(f"Font is NOT registered: {font_path_or_name}")


@app.command(
    name="is_font_registered",
    help="Check if a font file is registered to be used for captions.",
)
def is_font_registered_cmd(
    font_path_or_name: Annotated[
        str,
        typer.Argument(
            ...,
            help="The path of the font file or the font's name to be checked.",
            show_default=False,
        ),
    ],
):
    font_name = is_font_registered(font_path_or_name)
    if font_name:
        typer.echo(f"Font is registered: {font_name}")
    else:
        typer.echo(f"Font is NOT registered: {font_path_or_name}")


@app.command(
    name="version",
    help="Show the version of Captametropolis.",
)
def version_cmd():
    typer.echo(f"Captametropolis v{__version__}")


def main():
    app()


if __name__ == "__main__":
    main()



import ctypes
import os
import shutil
import subprocess as sp
import sys
import winreg as wr

from fontTools.ttLib import TTFont
from lxml import etree as ET

from .errors import FontNotRegisteredError

_IMGMGCK_DOCTYPE = """
<!DOCTYPE typemap [
  <!ELEMENT typemap (type)+>
  <!ELEMENT type (#PCDATA)>
  <!ELEMENT include (#PCDATA)>
  <!ATTLIST type name CDATA #REQUIRED>
  <!ATTLIST type fullname CDATA #IMPLIED>
  <!ATTLIST type family CDATA #IMPLIED>
  <!ATTLIST type foundry CDATA #IMPLIED>
  <!ATTLIST type weight CDATA #IMPLIED>
  <!ATTLIST type style CDATA #IMPLIED>
  <!ATTLIST type stretch CDATA #IMPLIED>
  <!ATTLIST type format CDATA #IMPLIED>
  <!ATTLIST type metrics CDATA #IMPLIED>
  <!ATTLIST type glyphs CDATA #REQUIRED>
  <!ATTLIST type version CDATA #IMPLIED>
  <!ATTLIST include file CDATA #REQUIRED>
]>
"""
_STANDARD_FONTS_DIR = os.path.join(
    os.path.abspath(os.path.dirname(__file__)), "assets", "fonts"
)


def is_admin():
    if os.name == "nt":
        try:
            return ctypes.windll.shell32.IsUserAnAdmin()
        except:
            return False
    elif os.name == "posix":
        return os.getuid() == 0
    else:
        return False


def require_admin(quiet_run: bool = False, verbose: bool = False):
    if verbose:
        print("Checking admin privileges...")
    if not is_admin():
        if verbose:
            print("WARNING: You need admin privileges to run this script.")

        if ".py" in sys.argv[0]:
            command = [sys.executable] + sys.argv
        else:
            command = sys.argv
        command[0] = os.path.abspath(os.path.normpath(command[0]))
        if os.name == "nt" and command[0].startswith("\\\\?\\"):
            command[0] = command[0].replace("\\\\?\\", "")

        if verbose:
            print(command)

        if os.name == "posix":
            os.execvp("sudo", ["sudo"] + command)
        else:
            params = " ".join(command[1:])
            result = ctypes.windll.shell32.ShellExecuteW(
                None,
                "runas",
                command[0],
                params,
                None,
                0 if quiet_run else 1,
            )
            if result <= 32:
                if result == 5:
                    print("ERROR: Access is denied.")
                raise ctypes.WinError(result)
        sys.exit(0)
    else:
        if verbose:
            print("Admin privileges granted.")


def is_local_transcription_available(verbose: bool = False):
    try:
        import whisper

        use_local_whisper = True
        if verbose:
            print("Using local whisper model...")
    except ImportError:
        use_local_whisper = False
        if verbose:
            print("Using OpenAI Whisper API...")

    return use_local_whisper


def ffmpeg_binary() -> str:
    return shutil.which("ffmpeg") or "unset"


def ffmpeg_installed() -> bool:
    return ffmpeg_binary() != "unset"


def imagemagick_directory() -> str:
    binary = None
    if os.name == "nt":
        try:
            key = wr.OpenKey(wr.HKEY_LOCAL_MACHINE, "SOFTWARE\\ImageMagick\\Current")
            binary = wr.QueryValueEx(key, "BinPath")[0]
            key.Close()
            return binary
        except:
            return "unset"
    else:
        try:
            result = sp.run(
                ["which", "convert"],
                capture_output=True,
                text=True,
                check=True,
            )
            convert_path = result.stdout.strip()

            if convert_path:
                imagemagick_dir = os.path.dirname(convert_path)
                return imagemagick_dir
            else:
                return "unset"
        except sp.CalledProcessError:
            return "unset"


def imagemagick_binary() -> str:
    def try_cmd(cmd):
        try:
            popen_params = {
                "stdout": sp.PIPE,
                "stderr": sp.PIPE,
                "stdin": sp.DEVNULL,
            }

            if os.name == "nt":
                popen_params["creationflags"] = 0x08000000

            proc = sp.Popen(cmd, **popen_params)  # type: ignore
            proc.communicate()
        except Exception as err:
            return False, err
        else:
            return True, None

    if os.name == "nt":
        imgmgck_directory = imagemagick_directory()
        if imgmgck_directory == "unset":
            return "unset"
        imgmgck_executable = (
            "magick.exe" if "ImageMagick-7" in imgmgck_directory else "convert.exe"
        )
        imgmgck_binary = os.path.join(imgmgck_directory, imgmgck_executable)
        if os.path.isfile(imgmgck_binary):
            return imgmgck_binary
        else:
            return "unset"
    elif try_cmd(["convert"])[0]:
        return "convert"
    else:
        return "unset"


def _get_font_info(font_path: str) -> dict:
    font = TTFont(font_path)

    name_table = font["name"]
    font_info = {}

    name_id_map = {
        1: "Font Family",
        2: "Font Subfamily",
        4: "Full Font Name",
        5: "Version",
        6: "Postscript Name",
        7: "Italic Angle",
    }

    for record in name_table.names:  # type: ignore
        name_id = record.nameID
        if name_id in name_id_map:
            name_type = name_id_map[name_id]
            font_info[name_type] = record.toUnicode()

    if "OS/2" in font:
        os2_table = font["OS/2"]
        weight = os2_table.usWeightClass  # type: ignore
        stretch = os2_table.usWidthClass  # type: ignore

        font_info["Weight"] = weight
        stretch_map = {
            1: "UltraCondensed",
            2: "ExtraCondensed",
            3: "Condensed",
            4: "SemiCondensed",
            5: "Normal",
            6: "SemiExpanded",
            7: "Expanded",
            8: "ExtraExpanded",
            9: "UltraExpanded",
        }
        font_info["Stretch"] = stretch_map[stretch].lower()

    return font_info


def register_font(fontpath: str, quiet_run: bool = False, verbose: bool = False) -> str:
    if not os.path.isfile(fontpath):
        fontpath = os.path.join(_STANDARD_FONTS_DIR, os.path.basename(fontpath))
    if not os.path.isfile(fontpath):
        raise FileNotFoundError(f"Font file not found: {fontpath}")

    font_container = os.path.join(imagemagick_directory(), "type-ghostscript.xml")
    if not os.path.isfile(font_container):
        raise FileNotFoundError("Font container not found")

    require_admin(quiet_run, verbose)

    font_info = _get_font_info(fontpath)

    tree = ET.parse(font_container)
    root = tree.getroot()

    for font in root.findall("type"):
        if font.attrib["fullname"] == font_info["Full Font Name"]:
            return font_info["Full Font Name"]

    new_font_type = ET.Element(
        "type",
        {
            "name": font_info["Full Font Name"],
            "fullname": font_info["Full Font Name"],
            "family": font_info["Font Family"],
            "style": "normal",
            "weight": str(font_info["Weight"]),
            "stretch": font_info["Stretch"],
            "format": "type1",
            "metrics": fontpath,
            "glyphs": fontpath,
        },
    )
    root.append(new_font_type)
    tree.write(
        font_container, encoding="utf-8", xml_declaration=True, doctype=_IMGMGCK_DOCTYPE
    )

    return font_info["Full Font Name"]


def is_font_registered(font_path_or_name: str) -> str | None:
    font_container = os.path.join(imagemagick_directory(), "type-ghostscript.xml")
    if not os.path.isfile(font_container):
        raise FileNotFoundError("Font container not found")

    if os.path.isfile(font_path_or_name):
        font_name = _get_font_info(font_path_or_name)["Full Font Name"]
    elif os.path.isfile(
        os.path.join(_STANDARD_FONTS_DIR, os.path.basename(font_path_or_name))
    ):
        font_name = _get_font_info(
            os.path.join(_STANDARD_FONTS_DIR, os.path.basename(font_path_or_name))
        )["Full Font Name"]
    else:
        font_name = font_path_or_name.removesuffix(".ttf")

    tree = ET.parse(font_container)
    root = tree.getroot()

    for font in root.findall("type"):
        if font.attrib["fullname"] == font_name:
            return font_name

    return None


def _get_font_path(font: str) -> tuple[str, str]:
    if not font.endswith(".ttf"):
        raise ValueError("Only TrueType fonts are currently supported")

    if os.path.isfile(font):
        injected_font_name = is_font_registered(font)
        if injected_font_name:
            return os.path.abspath(font), injected_font_name

        raise FontNotRegisteredError(font)

    font = os.path.join(_STANDARD_FONTS_DIR, os.path.basename(font))

    if not os.path.isfile(font):
        raise FileNotFoundError(f"Font '{font}' not found")

    injected_font_name = is_font_registered(font)
    if injected_font_name:
        return os.path.abspath(font), injected_font_name

    raise FontNotRegisteredError(font)


def unregister_font(
    font_path_or_name: str, quiet_run: bool = False, verbose: bool = False
) -> bool:
    font_container = os.path.join(imagemagick_directory(), "type-ghostscript.xml")
    if not os.path.isfile(font_container):
        raise FileNotFoundError("Font container not found")

    require_admin(quiet_run, verbose)

    if os.path.isfile(font_path_or_name):
        font_name = _get_font_info(font_path_or_name)["Full Font Name"]
    elif os.path.isfile(
        os.path.join(_STANDARD_FONTS_DIR, os.path.basename(font_path_or_name))
    ):
        font_name = _get_font_info(
            os.path.join(_STANDARD_FONTS_DIR, os.path.basename(font_path_or_name))
        )["Full Font Name"]
    else:
        font_name = font_path_or_name.removesuffix(".ttf")

    try:
        tree = ET.parse(font_container)
        root = tree.getroot()
        for font in root.findall("type"):
            if font.attrib["fullname"] == font_name:
                root.remove(font)
                tree.write(
                    font_container,
                    encoding="utf-8",
                    xml_declaration=True,
                    doctype=_IMGMGCK_DOCTYPE,
                )
                return True

        return False
    except Exception as e:
        raise e



class _CustomException(Exception):
    def __init__(self, message: str):
        self.__message__ = message
        super().__init__(self.__message__)

    @property
    def message(self):
        return self.__message__


class UtilityNotFoundError(_CustomException):
    def __init__(self, utility: str):
        super().__init__(
            f"Utility '{utility}' not found. Please make sure the utility is installed. For more information view the README.md file."
        )


class FontNotRegisteredError(_CustomException):
    def __init__(self, font: str):
        super().__init__(
            f"Font '{font}' is not registered. Please register the font using 'register_font_path()' before using it."
        )



from typing import Callable

def has_partial_sentence(text):
    words = text.split()
    if len(words) >= 2:
        prev_word = text.split()[-2].strip()
        if prev_word[-1] == ".":
            return True
    return False

def parse(
    segments: list[dict],
    fit_function: Callable,
    allow_partial_sentences: bool = False,
):
    captions = []
    caption = {
        "start": None,
        "end": 0,
        "words": [],
        "text": "",
    }

    # Merge words that are not separated by spaces
    for s, segment in enumerate(segments):
        for w, word in enumerate(segment["words"]):
            if w > 0 and word["word"][0] != " ":
                segments[s]["words"][w-1]["word"] += word["word"]
                segments[s]["words"][w-1]["end"] = word["end"]
                del segments[s]["words"][w]

    # Parse segments into captions that fit on the video
    for segment in segments:
        for word in segment["words"]:
            if caption["start"] is None:
                caption["start"] = word["start"]

            text = caption["text"]+word["word"]

            caption_fits = allow_partial_sentences or not has_partial_sentence(text)
            caption_fits = caption_fits and fit_function(text)

            if caption_fits:
                caption["words"].append(word)
                caption["end"] = word["end"]
                caption["text"] = text
            else:
                captions.append(caption)
                caption = {
                    "start": word["start"],
                    "end": word["end"],
                    "words": [word],
                    "text": word["word"],
                }

    captions.append(caption)

    return captions



#!/usr/bin/env python3

import os

from setuptools import find_packages, setup


def get_requirements():
    requirements_path = os.path.join(os.path.dirname(__file__), "requirements.txt")
    with open(requirements_path, "r", encoding="utf-8") as f:
        return f.read().splitlines()


def get_version():
    version_path = os.path.join(
        os.path.dirname(__file__), "captametropolis", "__version__.py"
    )
    with open(version_path, "r", encoding="utf-8") as f:
        for line in f:
            if line.startswith("__version__"):
                return line.split("=")[1].strip().strip('"')
    return "1.0.0"


setup(
    name="captametropolis",
    version=get_version(),
    packages=find_packages(),
    install_requires=get_requirements(),
    extras_require={
        "local": ["openai-whisper"],
    },
    package_data={
        "captametropolis": ["assets/**/*"],
    },
    include_package_data=True,
    url="https://github.com/AppSolves/captametropolis",
    license="MIT",
    author="Original: Unconventional Coding | This Fork: AppSolves",
    maintainer="AppSolves",
    maintainer_email="contact@appsolves.dev",
    author_email="unconventionalcoding@gmail.com",
    description="Add Automatic Captions to YouTube Shorts with AI",
    long_description=open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    entry_points={
        "console_scripts": [
            "captametropolis=captametropolis.__cli__:main",
        ],
    },
)
